import argparse
import io
import logging
import zipfile
from typing import Tuple

from panther_analysis_tool import cli_output
from panther_analysis_tool.backend.client import (
    BulkUploadParams,
    BulkUploadValidateStatusResponse,
)
from panther_analysis_tool.backend.client import Client as BackendClient
from panther_analysis_tool.backend.client import UnsupportedEndpointError
from panther_analysis_tool.zip_chunker import ZipArgs, analysis_chunks


def run(backend: BackendClient, args: argparse.Namespace) -> Tuple[int, str]:
    if backend is None or not backend.supports_bulk_validate():
        return 1, "Invalid backend. `validate` is only supported via API token"

    typed_args = ZipArgs.from_args(args)
    chunks = analysis_chunks(typed_args)
    buffer = io.BytesIO()

    with zipfile.ZipFile(buffer, "w", zipfile.ZIP_DEFLATED) as zip_out:
        for name in chunks[0].files:
            zip_out.write(name)

    buffer.seek(0, 0)
    params = BulkUploadParams(zip_bytes=buffer.read())

    try:
        result = backend.bulk_validate(params)
        if result.is_valid():
            return 0, f"{cli_output.success('Validation success')}"

        return 1, cli_output.multipart_error_msg(result, "Validation failed")
    except UnsupportedEndpointError as err:
        logging.debug(err)
        return 1, cli_output.warning("Your Panther instance does not support this feature")

    except BaseException as err:  # pylint: disable=broad-except
        return 1, cli_output.multipart_error_msg(
            BulkUploadValidateStatusResponse.from_json({"error": str(err)}), "Validation failed"
        )



"""
Panther Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import io
import json
from typing import Any, Dict
from unittest import TestCase, mock

from panther_analysis_tool.backend.client import (
    BulkUploadParams,
    DeleteDetectionsParams,
    TranspileFiltersParams,
    TranspileToPythonParams,
)
from panther_analysis_tool.backend.lambda_client import LambdaClient, LambdaClientOpts


class MockBoto:
    invoke: mock.MagicMock

    def __init__(self, invoke_returns: Any) -> None:
        self.invoke = mock.MagicMock(return_value=invoke_returns)


# use this for standard lambda responses
def _make_mock_response(http_status: int, payload: Dict[str, Any]) -> Dict[str, Any]:
    return {
        "Payload": io.BytesIO(json.dumps(payload).encode("utf-8")),
        "ResponseMetadata": {"HTTPStatusCode": http_status},
    }


# use this for wrapped "gateway" responses
def _make_mock_response_with_string_body(
    http_status: int, payload: Dict[str, Any]
) -> Dict[str, Any]:
    return _make_mock_response(
        http_status=http_status,
        payload={
            "body": json.dumps(payload),
            "headers": {},
            "statusCode": http_status,
            "multiValueHeaders": {},
        },
    )


class TestLambdaClient(TestCase):
    def test_init(self) -> None:
        with mock.patch("boto3.client", return_value=MockBoto(invoke_returns=None)):
            lc = LambdaClient(
                LambdaClientOpts(datalake_lambda="x", user_id="user", aws_profile=None)
            )
            self.assertEqual(lc._user_id, "user")

    def test_bulk_upload(self) -> None:
        mock_lambda_res = _make_mock_response_with_string_body(
            http_status=200,
            payload={
                "rules": {"total": 0, "new": 0, "modified": 0, "deleted": 0},
                "queries": {"total": 0, "new": 0, "modified": 0, "deleted": 0},
                "policies": {"total": 2, "new": 0, "modified": 0, "deleted": 0},
                "data_models": {"total": 0, "new": 0, "modified": 0, "deleted": 0},
                "lookup_tables": {"total": 0, "new": 0, "modified": 0, "deleted": 0},
                "global_helpers": {"total": 0, "new": 0, "modified": 0, "deleted": 0},
                "correlation_rules": {"total": 0, "new": 0, "modified": 0, "deleted": 0},
            },
        )

        with mock.patch("boto3.client", return_value=MockBoto(invoke_returns=mock_lambda_res)):
            lc = LambdaClient(
                LambdaClientOpts(datalake_lambda="x", user_id="user", aws_profile=None)
            )
            result = lc.bulk_upload(BulkUploadParams(zip_bytes=b""))

            self.assertEqual(result.data.policies.total, 2)
            self.assertEqual(result.data.rules.total, 0)

    def test_bulk_validate(self) -> None:
        with self.assertRaises(BaseException):
            lc = LambdaClient(
                LambdaClientOpts(datalake_lambda="x", user_id="user", aws_profile=None)
            )
            lc.bulk_validate(BulkUploadParams(zip_bytes=b"yo"))

    def test_delete_detections(self) -> None:
        mock_lambda_res = _make_mock_response_with_string_body(
            http_status=200, payload={"ids": ["1", "2"], "savedQueryNames": None}
        )

        with mock.patch("boto3.client", return_value=MockBoto(invoke_returns=mock_lambda_res)):
            lc = LambdaClient(
                LambdaClientOpts(datalake_lambda="x", user_id="user", aws_profile=None)
            )
            result = lc.delete_detections(
                DeleteDetectionsParams(
                    dry_run=True, ids=["1", "2", "3"], include_saved_queries=False
                )
            )

            self.assertEqual(["1", "2"], result.data.ids)
            self.assertEqual([], result.data.saved_query_names)

    def test_transpile_to_python(self) -> None:
        with self.assertRaises(BaseException):
            lc = LambdaClient(
                LambdaClientOpts(datalake_lambda="x", user_id="user", aws_profile=None)
            )
            lc.transpile_simple_detection_to_python(TranspileToPythonParams(data=[""]))

    def test_transpile_filters(self) -> None:
        with self.assertRaises(BaseException):
            lc = LambdaClient(
                LambdaClientOpts(datalake_lambda="x", user_id="user", aws_profile=None)
            )
            lc.transpile_filters(TranspileFiltersParams(data=[""]))



from typing import Any

from panther_analysis_tool.backend.client import (
    BackendCheckResponse,
    BackendResponse,
    BulkUploadParams,
    BulkUploadResponse,
    BulkUploadValidateStatusResponse,
)
from panther_analysis_tool.backend.client import Client as BackendClient
from panther_analysis_tool.backend.client import (
    DeleteDetectionsParams,
    DeleteSavedQueriesParams,
    FeatureFlagsParams,
    FeatureFlagsResponse,
    GenerateEnrichedEventParams,
    GenerateEnrichedEventResponse,
    GetRuleBodyParams,
    GetRuleBodyResponse,
    ListSchemasParams,
    MetricsParams,
    MetricsResponse,
    PerfTestParams,
    ReplayResponse,
    TestCorrelationRuleParams,
    TestCorrelationRuleResponse,
    TranspileFiltersParams,
    TranspileFiltersResponse,
    TranspileToPythonParams,
    UpdateSchemaParams,
)


class MockBackend(BackendClient):
    def async_bulk_upload(self, params: BulkUploadParams) -> BackendResponse[BulkUploadResponse]:  # type: ignore
        pass

    def bulk_upload(self, params: BulkUploadParams) -> BackendResponse[BulkUploadResponse]:  # type: ignore
        pass

    def check(self) -> BackendCheckResponse:  # type: ignore
        pass

    def list_schemas(self, params: ListSchemasParams) -> BackendResponse[Any]:  # type: ignore
        pass

    def update_schema(self, params: UpdateSchemaParams) -> BackendResponse[Any]:  # type: ignore
        pass

    def delete_saved_queries(self, params: DeleteSavedQueriesParams) -> BackendResponse[Any]:  # type: ignore
        pass

    def delete_detections(self, params: DeleteDetectionsParams) -> BackendResponse[Any]:  # type: ignore
        pass

    def supports_async_uploads(self) -> bool:  # type: ignore
        pass

    def get_rule_body(self, params: GetRuleBodyParams) -> BackendResponse[GetRuleBodyResponse]:  # type: ignore
        pass

    def transpile_simple_detection_to_python(  # type: ignore
        self, params: TranspileToPythonParams
    ) -> BackendResponse[Any]:
        pass

    def test_correlation_rule(  # type: ignore
        self, params: TestCorrelationRuleParams
    ) -> BackendResponse[TestCorrelationRuleResponse]:
        pass

    def transpile_filters(  # type: ignore
        self, params: TranspileFiltersParams
    ) -> BackendResponse[TranspileFiltersResponse]:
        pass

    def supports_bulk_validate(self) -> bool:  # type: ignore
        pass

    def bulk_validate(self, params: BulkUploadParams) -> BulkUploadValidateStatusResponse:  # type: ignore
        pass

    def supports_perf_test(self) -> bool:  # type: ignore
        pass

    def get_metrics(self, params: MetricsParams) -> BackendResponse[MetricsResponse]:  # type: ignore
        pass

    def run_perf_test(self, params: PerfTestParams) -> BackendResponse[ReplayResponse]:  # type: ignore
        pass

    def supports_enrich_test_data(self) -> bool:  # type: ignore
        pass

    def generate_enriched_event_input(  # type: ignore
        self, params: GenerateEnrichedEventParams
    ) -> BackendResponse[GenerateEnrichedEventResponse]:
        pass

    def feature_flags(self, params: FeatureFlagsParams) -> BackendResponse[FeatureFlagsResponse]:  # type: ignore
        pass



from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

PAT_VERSION = (this_directory / "VERSION").read_text().strip()

install_requires = [
    "gql",
    "aiohttp",
    "boto3",
    "dynaconf",
    "jsonpath-ng",
    "jsonschema",
    "nested_lookup",
    "packaging",
    "requests",
    "ruamel.yaml",
    "schema",
    "semver",
    "sqlfluff",
    "panther-core==0.11.2",
    "typing-extensions",
    "jsonlines",
]

setup(
    name="panther_analysis_tool",
    version=PAT_VERSION,
    packages=find_packages(),
    license="AGPL-3.0",
    description="Panther command line interface for writing, testing, and packaging policies/rules.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Panther Labs Inc",
    author_email="pypi@runpanther.io",
    url="https://github.com/panther-labs/panther_analysis_tool",
    download_url=f"https://github.com/panther-labs/panther_analysis_tool/archive/v{PAT_VERSION}.tar.gz",
    keywords=["Security", "CLI"],
    scripts=["bin/panther_analysis_tool", "bin/pat"],
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Security",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3.11",
    ],
    include_package_data=True,
)



from panther_analysis_tool.backend.client import BackendMultipartError


class BColors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

    @classmethod
    def wrap(cls, start: str, text: str) -> str:
        return f"{start}{text}{cls.ENDC}"


def bold(text: str) -> str:
    return BColors.wrap(BColors.BOLD, text)


def header(text: str) -> str:
    return BColors.wrap(BColors.HEADER, text)


def blue(text: str) -> str:
    return BColors.wrap(BColors.OKBLUE, text)


def cyan(text: str) -> str:
    return BColors.wrap(BColors.OKCYAN, text)


def success(text: str) -> str:
    return BColors.wrap(BColors.OKGREEN, text)


def warning(text: str) -> str:
    return BColors.wrap(BColors.WARNING, text)


def underline(text: str) -> str:
    return BColors.wrap(BColors.UNDERLINE, text)


def failed(text: str) -> str:
    return BColors.wrap(BColors.FAIL, text)


def multipart_error_msg(result: BackendMultipartError, msg: str) -> str:
    return_str = "\n-----\n"

    if result.has_error():
        return_str += f"{bold('Error')}: {result.get_error()}\n-----\n"

    for issue in result.get_issues():
        if issue.path and issue.path != "":
            return_str += f"{bold('Path')}: {issue.path}\n"

        if issue.error_message and issue.error_message != "":
            return_str += f"{bold('Error')}: {issue.error_message}\n"

        return_str += "-----\n"

    return f"{return_str}\n{failed(msg)}"



import os
import tempfile
from typing import Dict, Final

from schema import Schema

from panther_analysis_tool.schemas import (
    CORRELATION_RULE_SCHEMA,
    DATA_MODEL_SCHEMA,
    DERIVED_SCHEMA,
    GLOBAL_SCHEMA,
    LOOKUP_TABLE_SCHEMA,
    PACK_SCHEMA,
    POLICY_SCHEMA,
    RULE_SCHEMA,
    SAVED_QUERY_SCHEMA,
    SCHEDULED_QUERY_SCHEMA,
)

PACKAGE_NAME: Final = "panther_analysis_tool"

VERSION_STRING: Final = "0.55.1"

CONFIG_FILE = ".panther_settings.yml"
DATA_MODEL_LOCATION = "./data_models"
HELPERS_LOCATION = "./global_helpers"
LUTS_LOCATION = "./lookup_tables"
DATA_MODEL_PATH_PATTERN = "*data_models*"
LUTS_PATH_PATTERN = "*lookup_tables*"
HELPERS_PATH_PATTERN = "*/global_helpers"
PACKS_PATH_PATTERN = "*/packs"
POLICIES_PATH_PATTERN = "*policies*"
QUERIES_PATH_PATTERN = "*queries*"
RULES_PATH_PATTERN = "*rules*"
TMP_HELPER_MODULE_LOCATION = os.path.join(tempfile.gettempdir(), "panther-path", "globals")


class AnalysisTypes:
    DATA_MODEL = "datamodel"
    DETECTION = "detection"
    GLOBAL = "global"
    LOOKUP_TABLE = "lookup_table"
    PACK = "pack"
    POLICY = "policy"
    SAVED_QUERY = "saved_query"
    SCHEDULED_QUERY = "scheduled_query"
    RULE = "rule"
    DERIVED = "derived"
    SCHEDULED_RULE = "scheduled_rule"
    SIMPLE_DETECTION = "simple_detection"
    CORRELATION_RULE = "correlation_rule"


# The UserID is required by Panther for some API calls, but we have no way of
# acquiring it, and it isn't used for anything. This is a valid UUID used by the
# Panther deployment tool to indicate this action was performed automatically.
PANTHER_USER_ID = "00000000-0000-4000-8000-000000000000"

RESERVED_FUNCTIONS = (
    "alert_context",
    "dedup",
    "description",
    "destinations",
    "reference",
    "runbook",
    "severity",
    "title",
)

VALID_SEVERITIES = ["INFO", "LOW", "MEDIUM", "HIGH", "CRITICAL"]

SCHEMAS: Dict[str, Schema] = {
    AnalysisTypes.DATA_MODEL: DATA_MODEL_SCHEMA,
    AnalysisTypes.GLOBAL: GLOBAL_SCHEMA,
    AnalysisTypes.LOOKUP_TABLE: LOOKUP_TABLE_SCHEMA,
    AnalysisTypes.PACK: PACK_SCHEMA,
    AnalysisTypes.POLICY: POLICY_SCHEMA,
    AnalysisTypes.SAVED_QUERY: SAVED_QUERY_SCHEMA,
    AnalysisTypes.SCHEDULED_QUERY: SCHEDULED_QUERY_SCHEMA,
    AnalysisTypes.RULE: RULE_SCHEMA,
    AnalysisTypes.DERIVED: DERIVED_SCHEMA,
    AnalysisTypes.SCHEDULED_RULE: RULE_SCHEMA,
    AnalysisTypes.CORRELATION_RULE: CORRELATION_RULE_SCHEMA,
}

SET_FIELDS = [
    "LogTypes",
    "PackIDs",
    "OutputIds",
    "SummaryAttributes",
    "Suppressions",
    "Tags",
]

BACKEND_FILTERS_ANALYSIS_SPEC_KEY = "_backend_filters"


class ReplayStatus:
    DONE = "DONE"
    CANCELED = "CANCELED"
    ERROR_EVALUATION = "ERROR_EVALUATION"
    ERROR_COMPUTATION = "ERROR_COMPUTATION"
    EVALUATION_IN_PROGRESS = "EVALUATION_IN_PROGRESS"
    COMPUTATION_IN_PROGRESS = "COMPUTATION_IN_PROGRESS"


ENABLE_CORRELATION_RULES_FLAG = "EnableCorrelationRules"



"""
Panther Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import json
from copy import deepcopy
from unittest import TestCase

from panther_analysis_tool.immutable import (
    ImmutableCaseInsensitiveDict,
    ImmutableList,
    json_encoder,
)


class TestImmutableDict(TestCase):
    def setUp(self) -> None:
        self.initial_dict = {"t": 10, "a": [{"b": 1, "c": 2}], "d": {"e": {"f": True}}}
        self.immutable_dict = ImmutableCaseInsensitiveDict(self.initial_dict)

    def test_assignment_not_allowed(self) -> None:
        with self.assertRaises(TypeError):
            # pylint: disable=E1137
            self.immutable_dict["a"] = 1  # type: ignore

    def test_nested_assignment_not_allowed(self) -> None:
        with self.assertRaises(TypeError):
            # pylint: disable=E1137
            self.immutable_dict["d"]["e"]["f"] = False

    def test_original_dict_not_mutated(self) -> None:
        _ = self.immutable_dict["a"]
        self.assertEqual(self.initial_dict, self.immutable_dict._container)

    def test_raises_error_for_non_existent_key(self) -> None:
        with self.assertRaises(KeyError):
            _ = self.immutable_dict["a-non-existent-key"]

    def test_getitem(self) -> None:
        self.assertEqual(self.immutable_dict["t"], self.initial_dict["t"])

    def test_nested_access(self) -> None:
        self.assertEqual(self.immutable_dict["a"][0]["b"], 1)
        self.assertEqual(self.immutable_dict["d"]["e"]["f"], True)

    def test_equality(self) -> None:
        # Equality with dict
        self.assertEqual(self.immutable_dict, self.initial_dict)
        # Equality with another instance
        equal_dict = deepcopy(self.initial_dict)
        other_immutable_dict = ImmutableCaseInsensitiveDict(equal_dict)
        self.assertEqual(other_immutable_dict, self.immutable_dict)

    def test_shallow_copy(self) -> None:
        self.assertEqual(self.immutable_dict._container, self.initial_dict)
        self.assertIsNot(self.immutable_dict._container, self.initial_dict)

    def test_get(self) -> None:
        self.assertIsInstance(self.immutable_dict.get("d"), ImmutableCaseInsensitiveDict)
        self.assertIsInstance(self.immutable_dict.get("a"), ImmutableList)

    def test_ensure_immutable(self) -> None:
        initial_dict = {
            "a": [[1, 2], [3, 4]],
            "b": {"c": {"d": 1}},
            "t": 10,
            "e": {"f": [{"g": 90}]},
        }
        immutable_dict = ImmutableCaseInsensitiveDict(initial_dict)
        # List of lists with immutable elements
        self.assertIsInstance(immutable_dict["a"], ImmutableList)
        self.assertIsInstance(immutable_dict["a"][0], ImmutableList)
        self.assertEqual(immutable_dict["a"][0][1], 2)
        # Two-level nested dictionary
        self.assertIsInstance(immutable_dict["b"], ImmutableCaseInsensitiveDict)
        self.assertIsInstance(immutable_dict["b"]["c"], ImmutableCaseInsensitiveDict)
        self.assertEqual(immutable_dict["b"]["c"]["d"], 1)
        # Plain immutable object at top-level
        self.assertIsInstance(immutable_dict["t"], int)
        self.assertEqual(immutable_dict["t"], 10)
        # Two-level dictionary with nested list as value
        self.assertIsInstance(immutable_dict["e"]["f"], ImmutableList)
        self.assertIsInstance(immutable_dict["e"]["f"][0], ImmutableCaseInsensitiveDict)
        self.assertEqual(immutable_dict["e"]["f"][0]["g"], 90)

    def test_copy(self) -> None:
        initial_dict = {"a": True, "b": {"c": {"e": True, "d": [{"g": False}]}}}
        immutable_dict = ImmutableCaseInsensitiveDict(initial_dict)
        dict_copy = immutable_dict.copy()
        self.assertIsNot(dict_copy, initial_dict)
        self.assertIsNot(dict_copy["b"], initial_dict["b"])
        self.assertEqual(dict_copy["b"], initial_dict["b"])
        self.assertIsInstance(dict_copy["b"], dict)
        self.assertIsInstance(dict_copy["b"]["c"], dict)
        self.assertIsInstance(dict_copy["b"]["c"]["d"], list)
        self.assertIsNot(dict_copy["b"]["c"]["d"], initial_dict["b"]["c"]["d"])  # type: ignore
        dict_copy["h"] = False
        self.assertIs(dict_copy["h"], False)


class TestImmutableList(TestCase):
    def setUp(self) -> None:
        self.initial_list = ["a", "b", "c"]
        self.immutable_list = ImmutableList(self.initial_list)

    def test_raises_error_on_non_existent_index(self) -> None:
        with self.assertRaises(IndexError):
            _ = self.immutable_list[10]

    def test_assignment_not_allowed(self) -> None:
        with self.assertRaises(TypeError):
            # pylint: disable=E1137
            self.immutable_list[0] = "d"  # type: ignore

    def test_getitem(self) -> None:
        self.assertEqual(self.immutable_list[0], self.initial_list[0])

    def test_equality(self) -> None:
        # List
        self.assertEqual(self.initial_list, self.immutable_list)
        # Tuple
        self.assertEqual(tuple(self.initial_list), self.immutable_list)
        # Same class
        self.assertEqual(ImmutableList(self.initial_list.copy()), self.immutable_list)

    def test_shallow_copy(self) -> None:
        self.assertEqual(list(self.immutable_list._container), self.initial_list)
        self.assertIsNot(self.immutable_list._container, self.initial_list)

    def test_ensure_immutable(self) -> None:
        initial_list = [[1, 2], [3, 4], {"a": {"b": 1}}]
        immutable_list = ImmutableList(initial_list)
        self.assertIsInstance(immutable_list[0], ImmutableList)
        self.assertIsInstance(immutable_list[2], ImmutableCaseInsensitiveDict)
        self.assertIsInstance(immutable_list[2]["a"], ImmutableCaseInsensitiveDict)

    def test_iteration_returns_immutable_objects(self) -> None:
        data = [{"b": {"c": True}}, {"d": {"c": False}}]
        immutable_list = ImmutableList(data)
        iterated_list = list(immutable_list)
        self.assertIsNot(iterated_list[0], data[0])
        self.assertIsNot(iterated_list[1], data[1])
        self.assertIsNot(iterated_list[0]["b"], data[0]["b"])
        self.assertIsInstance(iterated_list[0]["b"], ImmutableCaseInsensitiveDict)

    def test_copy(self) -> None:
        initial_list = [{"b": {"c": True}}, {"d": {"c": False}}]
        immutable_list = ImmutableList(initial_list)
        list_copy = immutable_list.copy()
        self.assertIsNot(list_copy, initial_list)
        self.assertIsNot(list_copy[0], initial_list[0])
        self.assertEqual(list_copy[0], initial_list[0])
        self.assertIsInstance(list_copy[0], dict)
        list_copy.append({"e": False})
        self.assertIs(list_copy[-1]["e"], False)


class TestImmutableNestedList(TestCase):
    def setUp(self) -> None:
        self.initial_dict = {"a": [1, 2]}
        self.immutable_dict = ImmutableCaseInsensitiveDict(self.initial_dict)

    def test_assignment_not_allowed(self) -> None:
        with self.assertRaises(TypeError):
            self.immutable_dict["a"][0] = 100

    def test_original_dict_not_mutated(self) -> None:
        _ = self.immutable_dict["a"][0]
        self.assertEqual(self.initial_dict, self.immutable_dict._container)

    def test_raises_error_for_non_existent_index(self) -> None:
        with self.assertRaises(IndexError):
            _ = self.immutable_dict["a"][2]


class TestJSONSerialization(TestCase):
    def test_immutable_list(self) -> None:
        initial_list = [1, 2, 3]
        immutable_list = ImmutableList(initial_list)
        self.assertEqual(json.dumps(initial_list), json.dumps(immutable_list, default=json_encoder))

    def test_immutable_dict(self) -> None:
        initial_dict = {"a": [1, 2, 3], "b": {"c": True}}
        immutable_dict = ImmutableCaseInsensitiveDict(initial_dict)
        self.assertEqual(json.dumps(initial_dict), json.dumps(immutable_dict, default=json_encoder))

    def test_raises_type_error_for_nonserializable_object(self) -> None:
        with self.assertRaises(TypeError):
            json.dumps({"test_case": TestCase}, default=json_encoder)


class TestCaseInsensitiveLookup(TestCase):
    def setUp(self) -> None:
        self.data = {
            "Content-Encoding": "gzip",
            "ACCEPT": "*/*",
            "X-Forwarded-For": "10.0.0.1, 10.0.0.2",
            "Request": {"HTTP_version": "1.1", "query": "p=1&r=2"},
        }
        self.case_insensitive_dict = ImmutableCaseInsensitiveDict(self.data)

    def test_membership_check(self) -> None:
        self.assertIn("accept", self.case_insensitive_dict)
        self.assertIn("Accept", self.case_insensitive_dict)
        self.assertIn("ACCEPT", self.case_insensitive_dict)
        self.assertIn("CONTENT-Encoding", self.case_insensitive_dict)
        self.assertIn("REQUEST", self.case_insensitive_dict)
        self.assertNotIn("Content-Length", self.case_insensitive_dict)
        self.assertIn("http_version", self.case_insensitive_dict["request"])

    def test_immutable_return_value(self) -> None:
        self.assertIsInstance(self.case_insensitive_dict["request"], ImmutableCaseInsensitiveDict)
        self.assertEqual(self.case_insensitive_dict["request"], self.data["Request"])
        self.assertEqual(self.case_insensitive_dict["request"]["Query"], "p=1&r=2")

    def test_getitem(self) -> None:
        self.assertEqual(self.case_insensitive_dict["accept"], self.data["ACCEPT"])
        self.assertEqual(
            self.case_insensitive_dict["CONTENT-ENCODING"], self.data["Content-Encoding"]
        )
        self.assertEqual(self.case_insensitive_dict["request"]["http_version"], "1.1")
        with self.assertRaises(KeyError):
            _ = self.case_insensitive_dict["Unknown-key"]

    def test_get(self) -> None:
        self.assertEqual(
            self.case_insensitive_dict.get("x-forwarded-for"), self.data["X-Forwarded-For"]
        )
        self.assertEqual(
            self.case_insensitive_dict.get("X-FORWARDED-FOR"), self.data["X-Forwarded-For"]
        )
        self.assertIsNone(self.case_insensitive_dict.get("Unknown-key"))
        self.assertEqual(self.case_insensitive_dict.get("Unknown-key", 1), 1)

    def test_original_keys_on_iteration(self) -> None:
        self.assertListEqual(list(self.data), list(self.case_insensitive_dict))
        self.assertListEqual(list(self.data.keys()), list(self.case_insensitive_dict.keys()))
        self.assertListEqual(list(self.data.items()), list(self.case_insensitive_dict.items()))

    def test_on_conflict_first_occurrence_returned(self) -> None:
        data = {"X-Forwarded-For": "10.0.0.1, 10.0.0.2", "X-FORWARDED-FOR": "10.0.0.3, 10.0.0.4"}
        case_insensitive_dict = ImmutableCaseInsensitiveDict(data)
        self.assertEqual(case_insensitive_dict["X-FORWARDED-FOR"], "10.0.0.3, 10.0.0.4")
        self.assertEqual(case_insensitive_dict["x-forwarded-for"], data["X-Forwarded-For"])

    def test_on_empty_container_raises_key_error(self) -> None:
        case_insensitive_dict = ImmutableCaseInsensitiveDict({})
        with self.assertRaisesRegex(KeyError, "non_existing_key"):
            _ = case_insensitive_dict["non_existing_key"]

    def test_incremental_map_build(self) -> None:
        # Normally protected attributes should not participate in testing,
        # however this is critical functionality, so we maximize the test coverage.
        self.assertIn("CONTENT-ENCODING", self.case_insensitive_dict)
        self.assertEqual(
            self.case_insensitive_dict._case_insensitive_keymap,
            {"content-encoding": "Content-Encoding"},
        )
        self.assertIn("X-FORWARDED-FOR", self.case_insensitive_dict)
        self.assertEqual(
            self.case_insensitive_dict._case_insensitive_keymap,
            {
                "content-encoding": "Content-Encoding",
                "accept": "ACCEPT",
                "x-forwarded-for": "X-Forwarded-For",
            },
        )
        # Ensure existing keys in the key map are returned
        self.assertIn("CONTENT-ENCODING", self.case_insensitive_dict)
        self.assertIn("REQUEST", self.case_insensitive_dict)
        self.assertEqual(
            self.case_insensitive_dict._case_insensitive_keymap,
            {
                "content-encoding": "Content-Encoding",
                "accept": "ACCEPT",
                "x-forwarded-for": "X-Forwarded-For",
                "request": "Request",
            },
        )
        # Ensure after the keymap is fully built keys are successfully matched
        self.assertIn("X-FORWARDED-FOR", self.case_insensitive_dict)
        self.assertTrue(self.case_insensitive_dict._keymap_fully_built)



"""
Panther Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2023 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import re
import unittest
from argparse import Namespace
from pathlib import Path

from panther_analysis_tool.main import check_packs

FIXTURES_PATH = Path(__file__).parents[2] / "fixtures/check-packs"


class TestCheckPacks(unittest.TestCase):
    _bad_packs: dict = None

    def get_bad_packs(self) -> dict:
        if self._bad_packs is not None:
            return self._bad_packs

        args = Namespace(path=FIXTURES_PATH / "missing-dependencies")
        exit_code, res = check_packs(args)

        assert exit_code == 1
        res = res.split("\n")
        self._bad_packs = {}
        current_key = ""
        for line in res[1:]:
            if line.startswith(" ") or not line:
                continue
            if line.startswith("\t"):
                self._bad_packs[current_key].append(line.strip())
                continue
            current_key = Path(line.strip()).name
            self._bad_packs[current_key] = []

        return self._bad_packs

    def test_fixtures(self) -> None:
        args = Namespace(path=FIXTURES_PATH / "packless-rule")
        exit_code, res = check_packs(args)

        assert exit_code == 1
        assert "Test.Missing" in res

    def test_missing_global(self) -> None:
        bad_packs = self.get_bad_packs()
        assert "missing_global.yml" in bad_packs
        missing_items = bad_packs["missing_global.yml"]
        assert "a_helper" in missing_items

    def test_missing_data_model(self) -> None:
        bad_packs = self.get_bad_packs()
        assert "missing_datamodel.yml" in bad_packs
        missing_items = bad_packs["missing_datamodel.yml"]
        assert "panther" in missing_items
        assert "Standard.AWS.CloudTrail" in missing_items

    def test_missing_query(self) -> None:
        bad_packs = self.get_bad_packs()
        assert "missing_query.yml" in bad_packs
        missing_items = bad_packs["missing_query.yml"]
        assert "A Test Query" in missing_items

    def test_missing_subrules(self) -> None:
        bad_packs = self.get_bad_packs()
        assert "missing_subrules.yml" in bad_packs
        missing_items = bad_packs["missing_subrules.yml"]
        assert "AWS.CloudTrail.IaaS" in missing_items
        assert "GitHub.CICD" in missing_items

    def test_empty_dir(self) -> None:
        current_dir = os.path.dirname(os.path.realpath(__file__))
        print(current_dir)
        args = Namespace(path=current_dir)
        exit_code, res = check_packs(args)

        assert exit_code == 0
        assert res == "Looks like packs are up to date"



"""
Panther Analysis Tool is a command line interface for writing,
testing, and packaging policies/rules.
Copyright (C) 2020 Panther Labs Inc

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import dataclasses
import io
import json
import logging
import os
from fnmatch import fnmatch
from typing import Any, Callable, Dict, Generator, Iterator, List, Optional, Tuple

from ruamel.yaml import YAML
from ruamel.yaml import parser as YAMLParser
from ruamel.yaml import scanner as YAMLScanner

from panther_analysis_tool.backend.client import BackendError
from panther_analysis_tool.backend.client import Client as BackendClient
from panther_analysis_tool.backend.client import (
    GetRuleBodyParams,
    TestCorrelationRuleParams,
    TranspileFiltersParams,
    TranspileToPythonParams,
)
from panther_analysis_tool.constants import (
    BACKEND_FILTERS_ANALYSIS_SPEC_KEY,
    DATA_MODEL_PATH_PATTERN,
    HELPERS_PATH_PATTERN,
    LUTS_PATH_PATTERN,
    PACKS_PATH_PATTERN,
    POLICIES_PATH_PATTERN,
    QUERIES_PATH_PATTERN,
    RULES_PATH_PATTERN,
    VERSION_STRING,
    AnalysisTypes,
)
from panther_analysis_tool.util import is_simple_detection


class ClassifiedAnalysis:
    def __init__(self, file_name: str, dir_name: str, analysis_spec: Dict[str, Any]):
        self.file_name = file_name
        self.dir_name = dir_name
        self.analysis_spec = analysis_spec

    def is_deprecated(self) -> bool:
        display_name = self.analysis_spec["DisplayName"]
        description = self.analysis_spec.get("Description", "")
        if "deprecated" in display_name.lower() or "deprecated" in description.lower():
            return True

        tags = {tag.lower() for tag in self.analysis_spec.get("Tags", [])}
        if "deprecated" in tags:
            return True
        return False


@dataclasses.dataclass
class ClassifiedAnalysisContainer:
    """Contains all classified analysis specs"""

    data_models: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)
    globals: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)
    detections: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)
    simple_detections: List[ClassifiedAnalysis] = dataclasses.field(
        init=False, default_factory=list
    )
    queries: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)
    lookup_tables: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)
    packs: List[ClassifiedAnalysis] = dataclasses.field(init=False, default_factory=list)

    def _self_as_list(self) -> List[List[ClassifiedAnalysis]]:
        return [
            self.data_models,
            self.globals,
            self.detections,
            self.simple_detections,
            self.queries,
            self.lookup_tables,
            self.packs,
        ]

    def empty(self) -> bool:
        return all(len(l) == 0 for l in self._self_as_list())

    def apply(
        self,
        func: Callable[[List[ClassifiedAnalysis]], List[ClassifiedAnalysis]],
    ) -> "ClassifiedAnalysisContainer":
        container = ClassifiedAnalysisContainer()
        container.data_models = func(self.data_models)
        container.globals = func(self.globals)
        container.detections = func(self.detections)
        container.simple_detections = func(self.simple_detections)
        container.queries = func(self.queries)
        container.lookup_tables = func(self.lookup_tables)
        container.packs = func(self.packs)
        return container

    def items(self) -> Generator[ClassifiedAnalysis, None, None]:
        for analysis_list in self._self_as_list():
            for classified in analysis_list:
                yield classified

    def add_classified_analysis(
        self, analysis_type: str, classified_analysis: ClassifiedAnalysis
    ) -> None:
        if is_simple_detection(classified_analysis.analysis_spec):
            self.simple_detections.append(classified_analysis)
        elif analysis_type in [
            AnalysisTypes.POLICY,
            AnalysisTypes.RULE,
            AnalysisTypes.SCHEDULED_RULE,
            AnalysisTypes.CORRELATION_RULE,
        ]:
            self.detections.append(classified_analysis)
        elif analysis_type == AnalysisTypes.DATA_MODEL:
            self.data_models.append(classified_analysis)
        elif analysis_type == AnalysisTypes.GLOBAL:
            self.globals.append(classified_analysis)
        elif analysis_type == AnalysisTypes.LOOKUP_TABLE:
            self.lookup_tables.append(classified_analysis)
        elif analysis_type == AnalysisTypes.PACK:
            self.packs.append(classified_analysis)
        elif analysis_type == AnalysisTypes.SAVED_QUERY:
            self.queries.append(classified_analysis)
        elif analysis_type == AnalysisTypes.SCHEDULED_QUERY:
            self.queries.append(classified_analysis)


def filter_analysis(
    analysis: List[ClassifiedAnalysis], filters: Dict[str, List], filters_inverted: Dict[str, List]
) -> List[ClassifiedAnalysis]:
    if filters is None:
        return analysis

    filtered_analysis = []
    for item in analysis:
        dir_name = item.dir_name
        file_name = item.file_name
        analysis_spec = item.analysis_spec
        if fnmatch(dir_name, HELPERS_PATH_PATTERN):
            logging.debug("auto-adding helpers file %s", os.path.join(file_name))
            filtered_analysis.append(ClassifiedAnalysis(file_name, dir_name, analysis_spec))
            continue
        if fnmatch(dir_name, DATA_MODEL_PATH_PATTERN):
            logging.debug("auto-adding data model file %s", os.path.join(file_name))
            filtered_analysis.append(ClassifiedAnalysis(file_name, dir_name, analysis_spec))
            continue
        match = True
        for key, values in filters.items():
            spec_value = analysis_spec.get(key, "")
            spec_value = spec_value if isinstance(spec_value, list) else [spec_value]
            if not set(spec_value).intersection(values):
                match = False
                break
        for key, values in filters_inverted.items():
            spec_value = analysis_spec.get(key, "")
            spec_value = spec_value if isinstance(spec_value, list) else [spec_value]
            if set(spec_value).intersection(values):
                match = False
                break

        if match:
            filtered_analysis.append(ClassifiedAnalysis(file_name, dir_name, analysis_spec))

    return filtered_analysis


def load_analysis_specs(
    directories: List[str], ignore_files: List[str]
) -> Iterator[Tuple[str, str, Any, Any]]:
    """Loads the analysis specifications from a file. It calls load_analysis_specs_ex
    and returns the spec_filename, relative_path, analysis_spec, and error to preserve
    historic callers.

    Args:
        directories: The relative path to Panther policies or rules.
        ignore_files: Files that Panther Analysis Tool should not process

    Yields:
        A tuple of the relative filepath, directory name, and loaded analysis specification dict.
    """
    for result in load_analysis_specs_ex(directories, ignore_files, roundtrip_yaml=False):
        yield result.spec_filename, result.relative_path, result.analysis_spec, result.error


def disable_all_base_detections(paths: List[str], ignore_files: List[str]) -> None:
    analysis_specs = list(load_analysis_specs_ex(paths, ignore_files, roundtrip_yaml=True))
    base_ids_to_disable = set()
    base_detection_key = "BaseDetection"
    rule_id_key = "RuleID"
    enabled_key = "Enabled"
    for analysis_spec_res in analysis_specs:
        spec: Dict[str, Any] = analysis_spec_res.analysis_spec
        base_id = spec.get(base_detection_key, "")
        if base_id == "":
            continue
        base_ids_to_disable.add(base_id)
    for base_detection_id in base_ids_to_disable:
        for analysis_spec_res in analysis_specs:
            rule: Dict[str, Any] = analysis_spec_res.analysis_spec
            if rule.get(rule_id_key, "") == base_detection_id:
                logging.info(
                    "Setting %s=False for %s", enabled_key, analysis_spec_res.spec_filename
                )
                rule[enabled_key] = False
                analysis_spec_res.serialize_to_file()


@dataclasses.dataclass
class LoadAnalysisSpecsResult:
    """The result of loading analysis specifications from a file."""

    spec_filename: str
    relative_path: str
    analysis_spec: Any
    yaml_ctx: YAML
    error: Exception

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        spec_filename: str,
        relative_path: str,
        analysis_spec: Any,
        yaml_ctx: YAML,
        error: Any,
    ):
        self.spec_filename = spec_filename
        self.relative_path = relative_path
        self.analysis_spec = analysis_spec
        self.yaml_ctx = yaml_ctx
        self.error = error

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LoadAnalysisSpecsResult):
            return NotImplemented

        # skipping yaml_ctx because it's not relevant to equality of content
        # in analysis spec files
        same_spec_filename = self.spec_filename == other.spec_filename
        same_relative_path = self.relative_path == other.relative_path
        same_analysis_spec = self.analysis_spec == other.analysis_spec
        same_error = self.error == other.error

        return all(
            [
                same_spec_filename,
                same_relative_path,
                same_analysis_spec,
                same_error,
            ]
        )

    # pylint: disable=no-else-return
    def analysis_id(self) -> str:
        """Returns the analysis ID for this analysis spec."""
        analysis_type = self.analysis_spec["AnalysisType"]
        if analysis_type in [
            AnalysisTypes.RULE,
            AnalysisTypes.SCHEDULED_RULE,
            AnalysisTypes.CORRELATION_RULE,
        ]:
            return self.analysis_spec["RuleID"]
        elif analysis_type == AnalysisTypes.POLICY:
            return self.analysis_spec["PolicyID"]

        raise ValueError(f"Unknown analysis type '{analysis_type}'")

    def analysis_type(self) -> str:
        """Returns the analysis type for this analysis spec."""
        return self.analysis_spec["AnalysisType"]

    # pylint: disable=line-too-long
    def __str__(self) -> str:
        return f"LoadAnalysisSpecsResult(spec_filename={self.spec_filename}, relative_path={self.relative_path}, analysis_spec={self.analysis_spec['AnalysisType']}, error={self.error})"

    def serialize_to_file(self) -> None:
        logging.debug("Writing analysis spec to %s", self.spec_filename)
        with open(self.spec_filename, "w", encoding="utf-8") as file:
            self.yaml_ctx.dump(self.analysis_spec, file)


def get_yaml_loader(roundtrip: bool) -> YAML:
    """Returns a YAML object with the correct settings for loading analysis specifications.

    Args:
        roundtrip: Whether or not the YAML parser should be roundtrip safe. Roundtrip safe YAML
            parser is not compatible with many PAT functions.
    """
    if not roundtrip:
        return YAML(typ="safe")

    # If we need to roundtrip, we have different requirements. Most use cases will not need
    # round-tripping. We only need a roundtrip safe YAML parser if we are going to update
    # the YAML files.
    yaml = YAML(typ="rt")
    yaml.indent(mapping=2, sequence=4, offset=2)
    yaml.preserve_quotes = True
    yaml.default_flow_style = False
    # allow very long lines to avoid unnecessary line changes
    yaml.width = 4096
    return yaml


# pylint: disable=too-many-locals
def load_analysis_specs_ex(
    directories: List[str], ignore_files: List[str], roundtrip_yaml: bool
) -> Iterator[LoadAnalysisSpecsResult]:
    """Loads the analysis specifications from a file. The _ex variant of this function returns
    a LoadAnalysisSpecsResult object that contains the yaml_ctx object that was used to load
    the specific YAML content. This allows us to roundtrip the YAML content back to the file with
    minimal changes.

    Args:
        directories: The relative path to Panther policies or rules.
        ignore_files: Files that Panther Analysis Tool should not process
        roundtrip_yaml: If True, roundtrip the YAML content back to the file with minimal changes. This
            is incompatible with most of PAT's YAML loading use-cases. However, if you need to modify
            YAML code, this is necessary.

    Yields:
        An instance of LoadAnalysisSpecsResult.
    """
    # setup a list of paths to ensure we do not import the same files
    # multiple times, which can happen when testing from root directory without filters
    ignored_normalized = ["dependabot.yml", "package.json", "package-lock.json"]
    for file in ignore_files:
        ignored_normalized.append(os.path.normpath(file))

    loaded_specs: List[Any] = []
    for directory in directories:
        for relative_path, _, file_list in os.walk(directory):
            # Skip hidden folders
            if (
                relative_path.split("/")[-1].startswith(".")
                and relative_path != "./"
                and relative_path != "."
            ):
                continue

            # If the user runs with no path args, filter to make sure
            # we only run folders with valid analysis files. Ensure we test
            # files in the current directory by not skipping this iteration
            # when relative_path is the current dir
            if directory in [".", "./"] and relative_path not in [".", "./"]:
                if not any(
                    (
                        fnmatch(relative_path, path_pattern)
                        for path_pattern in (
                            DATA_MODEL_PATH_PATTERN,
                            HELPERS_PATH_PATTERN,
                            LUTS_PATH_PATTERN,
                            RULES_PATH_PATTERN,
                            PACKS_PATH_PATTERN,
                            POLICIES_PATH_PATTERN,
                            QUERIES_PATH_PATTERN,
                        )
                    )
                ):
                    logging.debug("Skipping path %s", relative_path)
                    continue
            for filename in sorted(file_list):
                # Skip hidden files
                if filename.startswith("."):
                    continue
                spec_filename = os.path.abspath(os.path.join(relative_path, filename))
                # skip loading files that have already been imported
                if spec_filename in loaded_specs:
                    continue
                # Dont load files that are explictly ignored
                relative_name = os.path.normpath(os.path.join(relative_path, filename))
                if relative_name in ignored_normalized:
                    logging.info("ignoring file %s", relative_name)
                    continue
                loaded_specs.append(spec_filename)
                # setup yaml object
                yaml = get_yaml_loader(roundtrip=roundtrip_yaml)
                if fnmatch(filename, "*.y*ml"):
                    with open(spec_filename, "r", encoding="utf-8") as spec_file_obj:
                        try:
                            yield LoadAnalysisSpecsResult(
                                spec_filename=spec_filename,
                                relative_path=relative_path,
                                analysis_spec=yaml.load(spec_file_obj),
                                yaml_ctx=yaml,
                                error=None,
                            )
                        except (YAMLParser.ParserError, YAMLScanner.ScannerError) as err:
                            # recreate the yaml object and yield the error
                            yield LoadAnalysisSpecsResult(
                                spec_filename=spec_filename,
                                relative_path=relative_path,
                                analysis_spec=None,
                                yaml_ctx=yaml,
                                error=err,
                            )
                if fnmatch(filename, "*.json"):
                    with open(spec_filename, "r", encoding="utf-8") as spec_file_obj:
                        try:
                            yield LoadAnalysisSpecsResult(
                                spec_filename=spec_filename,
                                relative_path=relative_path,
                                analysis_spec=json.load(spec_file_obj),
                                yaml_ctx=yaml,
                                error=None,
                            )
                        except ValueError as err:
                            yield LoadAnalysisSpecsResult(
                                spec_filename=spec_filename,
                                relative_path=relative_path,
                                analysis_spec=None,
                                yaml_ctx=yaml,
                                error=err,
                            )


def to_relative_path(filename: str) -> str:
    cwd = os.getcwd()
    return os.path.relpath(filename, cwd)


# This function was generated in whole or in part by GitHub Copilot.
def get_simple_detections_as_python(
    specs: List[ClassifiedAnalysis], backend: Optional[BackendClient] = None
) -> List[ClassifiedAnalysis]:
    """Returns simple detections with transpiled Python."""
    enriched_specs = []
    if backend is not None:
        batch = [json.dumps(spec.analysis_spec, allow_nan=False) for spec in specs]
        try:
            params = TranspileToPythonParams(data=batch)
            response = backend.transpile_simple_detection_to_python(params)
            if response.status_code == 200:
                for i, result in enumerate(response.data.transpiled_python):
                    item = specs[i]
                    spec = item.analysis_spec
                    spec["body"] = result
                    enriched_specs.append(ClassifiedAnalysis(item.file_name, item.dir_name, spec))
            else:
                logging.warning(
                    "Error transpiling simple detection(s) to Python, skipping tests for simple detections."
                )
        except (BackendError, BaseException) as be_err:  # pylint: disable=broad-except
            logging.warning(
                "Error transpiling simple detection(s) to Python, skipping tests for simple detections:  %s",
                be_err,
            )
    else:
        logging.info("No backend client provided, skipping tests for simple detections.")
    return enriched_specs if enriched_specs else specs


def lookup_base_detection(the_id: str, backend: Optional[BackendClient] = None) -> Dict[str, Any]:
    """Attempts to lookup base detection via its id"""
    out: Dict[str, Any] = {}
    if backend is not None:
        try:
            params = GetRuleBodyParams(id=the_id)
            response = backend.get_rule_body(params)
            if response.status_code == 200:
                out["body"] = response.data.body
                out["tests"] = response.data.tests
            else:
                logging.warning(
                    "Unexpected error getting base detection, status code %s", response.status_code
                )
        except (BackendError, BaseException) as be_err:  # pylint: disable=broad-except
            logging.warning(
                "Error getting base detection %s: %s",
                the_id,
                be_err,
            )
    return out


def test_correlation_rule(
    spec: Dict[str, Any], backend: Optional[BackendClient] = None
) -> List[Dict[str, Any]]:
    out: List[Dict[str, Any]] = []
    # dont make network call if there's no tests to run, or if no backend
    if "Tests" not in spec or backend is None:
        return out
    try:
        yaml = get_yaml_loader(roundtrip=True)
        string_io = io.StringIO()
        yaml.dump(spec, stream=string_io)
        output_str = string_io.getvalue()
        string_io.close()
        resp = backend.test_correlation_rule(
            TestCorrelationRuleParams(
                yaml=output_str,
            )
        )
        out = resp.data.results
    except Exception as be_err:  # pylint: disable=broad-except
        logging.warning(
            "Error running tests remotely for correlation rule %s: %s",
            spec.get("RuleID", ""),
            be_err,
        )
    return out


def transpile_inline_filters(
    all_specs: ClassifiedAnalysisContainer, backend: Optional[BackendClient] = None
) -> None:
    """
    Transpiles all InlineFilters on detections from simple detection format to the backend format and saves the
    transpiled filter's JSON to the analysis spec under "_backend_filters".
    """
    inline_filters_key = "InlineFilters"

    # separate out all detections with inline filters
    # keep track of the detections without filters so the list can be used later
    all_detections_with_filters, detections, simple_detections = [], [], []
    for det in all_specs.detections:
        if inline_filters_key in det.analysis_spec:
            all_detections_with_filters.append(det)
        else:
            detections.append(det)
    for det in all_specs.simple_detections:
        if inline_filters_key in det.analysis_spec:
            all_detections_with_filters.append(det)
        else:
            simple_detections.append(det)

    # if there are no InlineFilters at all then we can just return
    if len(all_detections_with_filters) == 0:
        return

    if backend is not None:
        batch = [
            json.dumps(d.analysis_spec.get("InlineFilters"), allow_nan=False)
            for d in all_detections_with_filters
        ]
        try:
            params = TranspileFiltersParams(data=batch, pat_version=VERSION_STRING)
            response = backend.transpile_filters(params)

            if response.status_code == 200:
                # set the translated backend filters on the analysis spec
                for i, result in enumerate(response.data.transpiled_filters):
                    all_detections_with_filters[i].analysis_spec[
                        BACKEND_FILTERS_ANALYSIS_SPEC_KEY
                    ] = json.loads(result)
                # separate the simple detections from the other detections
                for det in all_detections_with_filters:
                    if is_simple_detection(det.analysis_spec):
                        simple_detections.append(det)
                    else:
                        detections.append(det)
                # replace the lists in all specs with the new specs
                all_specs.detections = detections
                all_specs.simple_detections = simple_detections
            else:
                logging.warning(
                    "Error transpiling InlineFilter(s), skipping InlineFilters during testing"
                )
        except (BackendError, BaseException) as be_err:  # pylint: disable=broad-except
            logging.warning(
                "Error transpiling InlineFilter(s), skipping InlineFilters during testing:  %s",
                be_err,
            )
    else:
        logging.info("No backend client provided, skipping InlineFilters during testing")
